#!/usr/bin/perl -w

package DataCash::CardInfo;

use FileHandle;
use Cwd;
use DataCash::Config;
use DataCash::XMLDocument;

$DataCash::CardInfo::VERSION = do { my @r = (q$Revision: 1.13 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };

sub new {

  my $type=shift;
  
  my $self={};
  
  my %params=@_;
  foreach my $key (keys %params) {
    if (grep /^$key$/,qw(debug logfh config datadir pan)) {
      $self->{$key}=$params{$key};
    }
  }
  
  bless $self,$type;
  
  if (!defined $self->{config} && defined $params{configfile} && -e $params{configfile}) {
    $self->{config}=new DataCash::Config($params{configfile});
  }
  
  if ($self->{datadir}) {
    $self->{datadir} =~ s|([^/])$|$1/|;
    $self->{fh_bins_path}=$self->{datadir}."CardInfo1.bin";
    $self->{fh_data_path}=$self->{datadir}."CardInfo2.bin";
  } else {
    my $fbase=$self->_filebase();
    $self->{fh_bins_path}=$fbase."/CardInfo1.bin";
    $self->{fh_data_path}=$fbase."/CardInfo2.bin";
  }
  
  unless ($self->{fh_bins}=new FileHandle($self->{fh_bins_path},"r")) {
    die "Cannot open CardInfo1.bin: $!";
  }
  
  unless ($self->{fh_data}=new FileHandle($self->{fh_data_path},"r")) {
    die "Cannot open CardInfo2.bin: $!";
  }
  
  binmode $self->{fh_bins};
  binmode $self->{fh_data};

  $self->_find if $self->{pan};

  return $self;
  
}

sub validation {

  # Get the DataCash::CardInfo object
  my $self = shift;

  # Get the DataCash::Request object
  $self->{request} = shift;

  my $ret;
  
  
  if ($self->{request}->get("Transaction.CardTxn")) {
  # Get the PAN before we try to find the binary details, in case
  # there are any non-numerics in it
    my $pan = $self->{request}->get("Transaction.CardTxn.Card.pan");
    return $self->do_failure("-5", "card number required") if (! $pan );
    if (exists $self->{pan} && defined $self->{pan} && $pan ne $self->{pan}) {
      return $self->do_failure("-5", "card number in document does not match cardinfo pan");
    }
    $self->{pan} = $pan;
    return $self->do_failure("-5", "card number required") if (! $self->{pan});

    $self->{pan}=~s/\D//g;
    
    # First, let's find the details from the binary file
    unless ($self->{binfound}) {
      $self->_log(5, "About to find the BIN details of this card");
      $self->_find();
    }
    
    # do credit card checks
    $ret = $self->_cccheck();

    if (ref($ret) eq 'DataCash::XMLDocument') {
      return $ret;
    }
  }
  
  # Currently disabled
#  if ($self->{request}->get("Transaction.ChequeTxn")) {
#    $self->_chequecheck();
#  }
  
  if ($self->{request}->get("Transaction.TxnDetails.merchantreference")) {
    $ret = $self->_refcheck();
    if (ref($ret) eq 'DataCash::XMLDocument') {
      return $ret;
    }
  } 
}

sub _cccheck {

  my $self=shift;
  
  # Check we have located the bin details.  If not, there ain't
  # much point in going any further.
  
  unless (defined $self->{binfound} && $self->{binfound} eq 'true') { return }

  $self->_log(2,"Continuing with validation as we have card details.");

  # check length
  my $expected_length = $self->digits;
  my $mode = $self->{request}->get("Transaction.TxnDetails.mode");

  # ficticious test card, let it through
  if ($expected_length == 0 and defined $mode and $mode =~ /test/i){
    return;
  }

  $self->_log(2,"Expected: $expected_length, Actual: ".length($self->{pan}));

  if ( $expected_length != 0 ) {
    return $self->do_failure("-26","card number too long") if( length ($self->{pan}) > $expected_length );
    return $self->do_failure("-26","card number too short") if ( length ($self->{pan}) < $expected_length);
  }
  
  # is it valid - pass luhn check?
  return $self->do_failure("-25","Bad checksum") if(! $self->validnumber());

  # check the expiry date
  return $self->do_failure("-5","Expiry date required") if(! $self->{request}->get("Transaction.CardTxn.Card.expirydate"));
  my $expiry = $self->{request}->get("Transaction.CardTxn.Card.expirydate");
  return $self->do_failure("-23","Invalid Expiry date") if $expiry !~ /\d{2}\/\d{2}/;
  
  # check issue number/startdate
  if(!defined $self->{request}->get("Transaction.CardTxn.Card.issuenumber") and ($self->issuenumber()) != 0){
    return $self->do_failure("-5","Issuenumber required");
  }
  elsif ($self->issuenumber() != 0) {
    my $issuenumber = $self->{request}->get("Transaction.CardTxn.Card.issuenumber");
    return $self->do_failure("-27","Bad issue") if( $issuenumber !~ /^\d+$/);
#    return $self->do_failure("-27","Bad issue") if( $issuenumber =~ /^\d{2}$/ and $self->issuenumber() == 1);
#    return $self->do_failure("-27","Bad issue") if( $issuenumber =~ /^\d{1}$/ and $self->issuenumber() == 2);
  }

  if((!defined ($self->{request}->get("Transaction.CardTxn.Card.startdate"))) && $self->startdate() != 0){
    return $self->do_failure("-5","Startdate required");
  }
  elsif (defined $self->{request}->get("Transaction.CardTxn.Card.startdate")
	 && $self->{request}->get("Transaction.CardTxn.Card.startdate")) {
    my $startdate = $self->{request}->get("Transaction.CardTxn.Card.startdate");
    return $self->do_failure("-28","Bad start date") if $startdate !~ m/^\d{2}\/\d{2}$/;
  }

  return;
}

sub _chequecheck {

  my $self=shift;

  if($self->get("Transaction.ChequeTxn.Cheque.sortcode")){
    my $sortcode = $self->get("Transaction.ChequeTxn.Cheque.sortcode");
    if ( $sortcode !~ /^\d{6}$/ ){
      return $self->do_failure("-42","Invalid sortcode");
    }
  }else{
    return $self->do_failure("-5","sortcode required");
  }
  
  if($self->get("Transaction.ChequeTxn.Cheque.accountnumber")){
    my $accountnumber = $self->get("Transaction.ChequeTxn.Cheque.accountnumber");
       if ( $accountnumber !~ /^\d{8}$/ ){
         return $self->do_failure("-44","Invalid accountnumber");
       }
  }else{
    return $self->do_failure("-5","accountnumber required");
  }
  
  if($self->get("Transaction.ChequeTxn.Cheque.chequenumber")){
    my $chequenumber = $self->get("Transaction.ChequeTxn.Cheque.chequenumber");
    if ( $chequenumber !~ /^\d{6}$/ ){
      return $self->do_failure("-41","Invalid chequenumber");
    }
  }else{
    return $self->do_failure("-5","chequenumber required");
  }
  
  if($self->get("Transaction.ChequeTxn.Cheque")){
    my ($cheque, $chtype) = $self->get("Transaction.ChequeTxn.Cheque");
    if (defined $chtype and ref($chtype) =~ /^HASH/) {
      my $chtype = $chtype->{type};
      if( $chtype !~ /^(business|personal)$/){
	return $self->do_failure("-40","Invalid cheque type");
      }
    }else{
      return $self->do_failure("-5","cheque type required");
    }
  }
}

sub _refcheck {
  
  my $self = shift;
  $self->{ref} = $self->{request}->get("Transaction.TxnDetails.merchantreference");

  if( length($self->{ref}) < 6 || length($self->{ref}) > 30 ){

    return $self->do_failure("-22","Invalid reference");

  }

}

sub scheme {
	my $self=shift;
	return $self->{scheme};
}

sub issuer {
	my $self=shift;
	return $self->{issuer};
}

sub digits {
	my $self=shift;
	return($self->{requirements} & 31);
}

sub country {
	my $self=shift;
	return $self->{country};
}

sub validnumber {
	my $self=shift;
	return $self->_luhn();
}

sub requirements {
	my $self=shift;
	
	my @reqs;
	
	push @reqs,("pan_".$self->digits());
	if ($self->cpass()) {
		push @reqs,("cpass");
	}
	if ($self->startdate()) {
		push @reqs,("startdate");
	}
	if ($self->issuenumber()==1) {
		push @reqs,("issue_1");
	} elsif ($self->issuenumber()==2) {
		push @reqs,("issue_2");
	}
	if ($self->nongbp()) {
		push @reqs,("supports_non-gbp");
	}
	if ($self->supportslid()) {
		push @reqs,("supports_lid");
	}
	
	
	return @reqs;
}

sub issuenumber {
	my $self=shift;
	return (($self->{requirements} & 192)/64);
}

sub startdate {
	my $self=shift;
	return (($self->{requirements} & 32)!=0);
}

sub cpass {
	my $self=shift;
	return (($self->{requirements} & 256)!=0);
}

sub nongbp {
	my $self=shift;
	return (($self->{requirements} & 1024)!=0);
}

sub supportslid {
	my $self=shift;
	return (($self->{requirements} & 512)!=0);
}

sub _filebase {
	my $self=shift;

		my $cwd=getcwd();
	unless (defined($self->{filebase})) {

		directory:
		foreach my $dir (@INC) {
			chdir ($dir);
			if (-d "DataCash") {
				chdir "DataCash";
				opendir DIR,"." or die "Serious drain-bamage: $!";
				if (grep /^CardInfo\d\.bin/,readdir(DIR)) {
					$self->{filebase}=getcwd();
					last directory;
				}
			}
		}
		closedir DIR;
	}
	
	chdir($cwd);
	return $self->{filebase};
}

sub _find {
  my $self=shift;
  my $offset=shift || 0;
  my $buffer;
  
  my $bin=substr($self->{pan},0,6);
  
  $self->{fh_bins}->seek($offset*16,0);
  $self->{fh_bins}->read($buffer,16);
  
  my ($start,$length,$left,$right,$data)=$self->_unbin($buffer);
  
  ##print "At $offset, BIN $start...\n";
  
  unless ($start) {
    return $self;
  }
  
  if ($bin>=($start+$length)) {
    ##print "Moving right (+$right) as $bin >= $start + $length\n";
    $self->_find($right+$offset);
  } else {
    if ($bin>=$start) {
      ##print "Got it!\n";
      $self->_log(1,"Got CardInfo details: $bin");
      $self->_data($data);
    } else {
      ##print "Moving left (+$left)\n";
      $self->_find($left+$offset);
    }
  }
}

sub _data {
	my $self=shift;
	my $offset=shift;
	my $scheme;
	my $issuer;
	my $country;
	my $requirements;

	# Note the fact that we have located the bin details
	$self->{binfound} = 'true';

	my $fh=$self->{fh_data};

	while (<$fh>) {
		chomp;
		unless (--$offset) {
			($scheme,$issuer,$country,$requirements)=split /\0/,$_;
			last;
		}
	}

	$self->{scheme}=$scheme;
	$self->{issuer}=$issuer;
	$self->{country}=$country;
	$self->{requirements}=$requirements;

	$self->_log(2,"Requirements $requirements | Scheme $scheme | Issuer " .
		    "$issuer | Country $country ");
	
	return $self;
}

sub _unbin {
	my $self=shift;
	my $buffer=shift;
	
	my $len=length($buffer);
	
	unless ($len==16) {
		$self->_log(1,"Error: Buffered read $len (not 16) octets in length : $buffer");
		##die "Error: Buffered read $len (not 16) octets in length";
		return undef;
	}

	my $start=_chunk($buffer,0,3);
	my $length=_chunk($buffer,3,1);
	my $left=_chunk($buffer,4,4);
	my $right=_chunk($buffer,8,4);
	my $data=_chunk($buffer,12,4);
	
	$self->_log(5, "$start,$length,$left,$right,$data");
	return ($start,$length,$left,$right,$data);
}

sub _chunk {
	my $data=shift;
	my $offset=shift;
	my $length=shift;
	
	my $template="x"x$offset."C"x$length;
	my @octets=unpack($template,$data);

	my $i=0;
	my $total=0;
	for (reverse @octets) {
		$total+=$_*(2**$i);
		$i+=8;
	}
	
	return $total;
}

sub _luhn {
  my $self=shift;
  my $pan=$self->{pan};
  
  $total = 0;
  while (length($pan)) {
    $total += chop($pan);
    if (length($pan)) {
      $total += (($d = chop($pan)) < 9 ? ($d * 2) % 9 : 9);
    }
  }
  $self->{validnumber}=!($total % 10);
}

sub _log {
	my $self=shift;
  my $whoami=ref($self);
	my $lvl=shift;
	
	if (defined $self->{config} && $self->{config}->{logging} >= $lvl) {
		foreach my $msg (@_) {
			if ((ref($self->{config}->{logfh})) eq "FileHandle") {
				my $fh=$self->{config}->{logfh};
				print $fh scalar(localtime(time))," [$$] $whoami $msg\n";
			} else {
	print STDERR " [$$] $whoami $msg\n";
			}
		}
	}
}

sub do_failure {

  my $self = shift;
  my $status = shift;
  my $reason = shift;
  my $merchantreference = $self->{request}->get("Transaction.TxnDetails.merchantreference");

  my $xml = new DataCash::XMLDocument();
  $xml->set("status", $status);
  $xml->set("reason", $reason);
  $xml->set("merchantreference", $merchantreference);

  return $xml;

}


sub DESTROY {
  my $self=shift;
  if (defined $self->{fh}) {
    $self->{fh}->close();
  }
}

1;
######################################################################!#########
__END__

=head1 NAME

DataCash::CardInfo - Validity checking of card details.

=head1 SYNOPSIS

   use DataCash::CardInfo;
   my $card=new DataCash::CardInfo(pan => "4444333322221111");
   unless ($card->validnumber()) {
      die "Not a valid card";
   }
   print $card->scheme(),"\n",$card->issuer(),"\n",$card->country(),"\n";

=head1 DESCRIPTION

DataCash::CardInfo is a replacement for the old CCCheck module. It uses a new
data storage method, and provides a more intuitive interface to rerieving data,
in order to make it easier to find out about cards.

=head1 PUBLIC METHODS

=head2 new()

The creator method, new(), invokes a new instance of DataCash::CardInfo. It is
passed the PAN, or cardnumber, and (optionally) a has of options. Any non-digit
characters in the PAN will be discarded, allowing you to use a number in the
form "4444-3333-2222-1111", if you I<really> must.

You must pass at least the first six digits of the PAN in order ot be able to
determine the type of the card. Only if the I<full> PAN is passed can the module
test that it fulfills the Luhn checksum algorithm.

Additionally, new() may be passed a hash of configuration options.

  * debug
  * logfh (which should be of type FileHandle)
  * datadir (which should be fully specified, and end with a trailing slash)

=head2 scheme()

The scheme() method indicates which card scheme the card belongs to. This will
be a string value such as "Switch", "Visa", "Visa Delta", etc.

=head2 issuer()

The issuer() method indicates (where possible) which financial institution issued
the card. This will be a string value such as "Royal Bank of Scotland", "Citibank", 
"Halifax", etc.

=head2 country()

The country() method will return (where possible) the ISO3166 three-letter country
code for the country in which the card was issued.

=head2 digits()

This method indicates how many digits long a card in this BIN range I<should> be.

=head2 validnumber()

This method indicates (by returning a boolean value) whether the supplied
card number passed the Luhn checksum algorithm.

=head2 requirements()

The requirements() method returns a list of parameters which must be passed and valid
in order to obtain an authorisation for the card. These may include:

  * startdate
  * issuenumber
  * cpass

These can also be checked for individually using the following accessor methods.

=head2 startdate

=head2 issuenumber

=head2 cpass

=head2 lidsupport

=head2 nongbp

=head1 ABOUT THE DATAFILES

DataCash::CardInfo gets its data from two datafiles, CardInfo1.bin and CardInfo2.bin,
which are located in the directory that the CardInfo module is installed. These datafiles
may be updated independently of the software module, by obtaining new datafile releases
from the DataCash web-site. However, you I<must> use the corresponding versions of these
files or else spurious results may be obtained by the software calls.

=head1 CHANGELOG

=head1 AUTHOR

Kas Robson (kxr@datacash.com), DataCash Ltd.

=head1 LICENSE
