#!/usr/bin/perl

package DataCash::Config;

use strict;
use FileHandle;
use Cwd;
use XML::Hash;

$DataCash::Config::VERSION = do { my @r = (q$Revision: 1.7 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };

sub new {
  my $type=shift;
  my $filename=shift;
  
  if (!defined $filename) {
    return undef;
  }
  
  my $self={};	
  bless $self,$type;
  
  unless ($filename) {
    $filename=$self->_defaultfile();
  }
  
  $self->{filename}=$filename;
  
  $self->_read();
  return $self;
}

sub _read {

  my $self=shift;
  
  my $fh;
  
  my $xml = new XML::Hash();

  unless ($fh=new FileHandle($self->{filename},"r")) {
    return "Cannot open $self->{filename}: $!";
  }
  
  my $xmldoc = "";
  while (<$fh>) {
    $xmldoc .= $_;
  }
  $xml->readxml($xmldoc);
  $xml->{xml} = $xml->{xml}->{Configuration};
  
  for ( qw(logfile logging host port timeout network setstrict blowfish proxy proxy_username proxy_password cacert_location) ) {
    $self->{$_} = $xml->get($_);
  }
  
  my $total = -1;
  if ($xml->is_array("Obscure.element")) {
    $total = $xml->is_array("Obscure.element") . "\n";
  }
  if ($total > 1) {
    $self->{obscure} = [];
    for (0 .. $total-1) {
      push @{$self->{obscure}}, $xml->get("Obscure.element[$_]");
    }
  }
  
  if (defined $self->{logfile}) {
    $self->{logfile} =~ s/[\"\']//g;
    unless ($self->{logfh}=(new FileHandle(">> $self->{logfile}"))) {
      return "Could not open $self->{logfile}: $!";
    }
    my $old_fh = select($self->{logfh});
    $| = 1;
    select($old_fh);
  }
  return $self;
}

sub _defaultfile {
	my $self=shift;

	my $cwd=getcwd();

	my $filebase;
	directory:
	foreach my $dir (@INC) {
		chdir ($dir);
		if (-d "DataCash") {
			chdir "DataCash";
			opendir DIR,"." or die "Serious drain-bamage: $!";
			if (grep /^Config\.cfg/,readdir(DIR)) {
				$filebase=getcwd();
				last directory;
			}
		}
	}
	closedir DIR;
	
	chdir($cwd);
	return "$filebase/Config.cfg";
}

sub DESTROY {
  my $self=shift;
  if (defined $self->{fh}) {
    $self->{logfh}->close();
  }
}

1;
######################################################################!#########
__END__

=head1 NAME

DataCash::Config - the interface for configuration parameters.

=head1 SYNOPSIS

   use DataCash::Config;
   my $cfg=new DataCash::Config("/etc/datacash.cfg");
   for (keys %{$cfg}) {
      print "$_: $cfg->{$_}\n";
   }

=head1 DESCRIPTION

The DataCash::Config package gives an object-oriented interface to an XML
configuration file, which stores certain parameters regarding the
APIs behaviour.

It must be constructed with a filename, this file will (if possible) be
loaded and read.

=head1 CONFIGURATION DIRECTIVES

=head2 logfile

The name of the logfile you wish to read from. This can be fully specified (eg.
/var/log/datacash.log) or you can give a relative path (eg. ./datacash.log), in
which case the file will be opened relative to where your code is executing -
I<not> relative to your configuration file.

=head2 logging

The logging level you wish to use, ranging from 1 (least verbose) to 5 (most
verbose).

=head2 host

This is where the transaction is to be sent.  Details of our servers are available
on the DataCash web site.

=head2 port

=head2 timeout

How long (in seconds) the API should wait for a response, before giving up.


=head2 network

What kind of connection you wish to utilise. If left unspecified, this will
default to HTTPS. Consult the perldoc for your version of DataCash::Net to
see what methods are enabled.

=head2 proxy

If you need to use a proxy, set this option, the value of which should be
the host and port of your proxy server.  If you do not use a proxy, simply
do not include this option in your configuration.

=head2 proxy_username

Should only be set if using a proxy, and your proxy required user authentication.

=head2 proxy_password

Should only be set if using a proxy, and your proxy required user authentication.

=head2 setstrict

Specify whether you wish the transactions to be rejected if the card details
fail validation.  If set to false, the transaction will be sent to the 
DataCash server regardless whether the API believes there to be information
missing or incorrectly specified.  If this value is not set in the 
configuration, it will default to false.

=head2 obscure

A list of nodes (seperated by spaces on a single line) which should be
obscured in your logfile.

=head2 cacert_location

The name of a file containing one or more root certificates to use to
verify the certificate of the host being connected to. If this option
is not present, the certificate will not be verified. It is highly
recommended that certificate verification is used for all live
transactions.

=head1 EXAMPLE

Here is an example of a simple configuration file.

<Configuration>
  <logfile>./datacash.log</logfile>
  <logging>2</logging>
  <Obscure>
    <element>Transaction.CardTxn.Card.pan</element>
    <element>Authentication.password</element>
    <element>Transaction.CardTxn.Card.Cv2Avs.cv2</element>
  </Obscure>
  <host>https://testserver.datacash.com/Transaction</host>
  <port>443</port>
  <timeout>120</timeout>
  <setstrict>true</setstrict>
  <proxy>http://our.proxy.server:10000</proxy>
  <cacert_location>./root-certs.pem</cacert_location>
</Configuration>



=head1 AUTHOR

Kas Robson (kxr@datacash.com), DataCash Ltd.
