#!/usr/bin/perl -w

package DataCash::XMLSSLAgent;

use strict;
use LWP;
use DataCash::XMLDocument;

$DataCash::XMLSSLAgent::VERSION = do { my @r = (q$Revision: 1.9 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };

sub new {

  my $package = shift;
  my $self = {};

  my %params=@_;
  
  foreach my $key (keys %params) {
    if (grep /^$key$/,qw(debug configfile)) {
      $self->{$key}=$params{$key};
    }
  }
  
  # Return undef if we ain't got a configfile
  if (!defined $self->{configfile}) {
    return undef;
  }

  # Create a Config object
  $self->{config} =new DataCash::Config($self->{configfile});
  
  bless $self, $package;
  $self->_log(3, "created new instance");
  $self->getErrorCodes();

  #Check certificate location
  if (defined $self->{config} &&
      defined $self->{config}->{cacert_location}) {
    my $location = $self->{config}->{cacert_location};
    if (! -e $location) {
      $self->_log(1, "Cannot find cacert_location: ".$location);
      die "Cannot find cacert_location: ".$location;
    }
    if (! -r $location) {
      $self->_log(1, "Cannot read cacert_location: ".$location);
      die "Cannot read cacert_location: ".$location;
    }
  }

  return $self;

}

sub send {

  my $self=shift;
  my $xmldoc = shift;

  if (defined $xmldoc->{failed_cardinfo} && 
      defined $self->{config} &&
      $xmldoc->{failed_cardinfo} eq 'true' &&
      defined $self->{config}->{setstrict} &&
      $self->{config}->{setstrict} eq 'true') {
      $self->_log(2,"Failed CardInfo Check, therefore not sending transaction.");
      return $xmldoc->{xmlresponse};
  }

  $xmldoc->set("UserAgent.Libraries.lib","DataCash::XMLSSLAgent",
	     version=>$DataCash::XMLSSLAgent::VERSION);

  if (!$xmldoc->get("UserAgent.language")) {
    $xmldoc->set("UserAgent.language","Perl",version=>$]);
  }
  
  if (defined $self->{config}->{network} && ($self->{config}->{network})=~/^break$/i) {
    die;
  } 
  elsif ( defined ($self->{config}->{network}) && 
	  $self->{config}->{network} !~ /^https$/i) {
    $self->_log(1,"does not support your connection type","falling back to HTTPS");
  }
  else {
    $self->_send_https($xmldoc);
  }
  
}

sub _send_https {
  
  my $self = shift;
  my $xmldoc = shift;

  my $host = $self->{config}->{host};	
  my $port = $self->{config}->{port};
  my $timeout = $self->{config}->{timeout};
  
  my $ua = new LWP::UserAgent;

  $ua->ssl_opts(
	SSL_verify_mode => 0x00,
	verify_hostname => 0,
  );

  $ENV{HTTPS_PROXY} = $self->{config}->{proxy} 
    if defined $self->{config}->{proxy} && $self->{config}->{proxy};
  $ENV{HTTPS_PROXY_USERNAME} = $self->{config}->{proxy_username} 
    if defined $self->{config}->{proxy_username} && $self->{config}->{proxy_username};
  $ENV{HTTPS_PROXY_PASSWORD} = $self->{config}->{proxy_password} 
    if defined $self->{config}->{proxy_password} && $self->{config}->{proxy_password};

  # Set the location of the root CA certificates if we have one.
  if (defined $self->{config}->{cacert_location} && $self->{config}->{cacert_location}) {
    $ENV{HTTPS_CA_FILE} = $self->{config}->{cacert_location};
  }

  my $header = new HTTP::Headers;
  my $httpreq = new HTTP::Request(POST=>$host,$header,$xmldoc->writexml("Request"));
  
  $self->_log(2,"sending data to $host");
  unless ($ua->is_protocol_supported("https")) {
    $self->_log(1,"does not support HTTPS!");
  }
  
  if (defined $timeout) {
    $ua->timeout($timeout);
  }
  
  my $start=time();
  my $httpres=$ua->request($httpreq);
  
  my $time=time()-$start;
  
  if ($httpres->is_success()) {

    $self->_log(2,"received response from $host");
    $self->_log(1,"took $time seconds");
    my $xmlresponse = new DataCash::XMLDocument();
    $xmlresponse->readxml($httpres->content());
    $xmlresponse->{xml} = $xmlresponse->{xml}->{Response};
    $self->_log(2, "Response:\n".$xmlresponse->writexml("Response"));
    return $xmlresponse;

  } 
  elsif ( $httpres->message() =~ /timeout/i ) {
    $self->_log(1,"Timeout occurred when sending transaction");
    return $self->make_response("-3", "Timeout");
  } elsif ( $httpres->message() =~ /SSL negotiation failed/i ) {
    $self->_log(1,"SSL Certificate verification failed");
    return $self->make_response("-5", "SSL certificate verification failed");
  } else {

    $self->_log(1,"Error contacting $host: Error Code " . $httpres->code);
    return $self->make_response("-5", "HTTPS Error Code " . $httpres->code . 
				" received, which indicates: " . 
				$self->{errorcodes}->{$httpres->code()});
  }
}

sub _log {

  my $self=shift;
  my $whoami=ref($self);
  my $lvl=shift;
  
  if (defined $self->{config}->{logging} && $self->{config}->{logging} >= $lvl) {
    foreach my $msg (@_) {
      if ((ref($self->{config}->{logfh})) eq "FileHandle") {
	my $fh=$self->{config}->{logfh};
	print $fh scalar(localtime(time))," [$$] $whoami $msg\n";
      } else {
	print STDERR " [$$] $whoami $msg\n";
      }
    }
  }

}

sub getErrorCodes {
  
  my $self = shift;
  $self->{errorcodes} = {
			 100 => 'Continue',
			 101 => 'Switching Protocols',
			 200 => 'OK',
			 201 => 'Created',
			 202 => 'Accepted',
			 203 => 'Non-Authoritative Information',
			 204 => 'No Content',
			 205 => 'Reset Content',
			 206 => 'Partial Content',
			 300 => 'Multiple Choices',
			 301 => 'Moved Permanently',
			 302 => 'Moved Temporarily',
			 303 => 'See Other',
			 304 => 'Not Modified',
			 305 => 'Use Proxy',
			 400 => 'Bad Request',
			 401 => 'Unauthorised',
			 402 => 'Payment Required',
			 403 => 'No Permission',
			 404 => 'Document Does Not Exist',
			 405 => 'Method Not Allowed',
			 406 => 'Not Acceptable',
			 407 => 'Proxy Authentication Required',
			 408 => 'Request Time-Out',
			 409 => 'Conflict',
			 410 => 'Gone',
			 411 => 'Length Required',
			 412 => 'Precondition Failed',
			 413 => 'Request Entity Too Large',
			 414 => 'Request-URI Too Large',
			 415 => 'Unsupported Media Type',
			 500 => 'TIMEOUT (or Error Running CGI Script)',
			 501 => 'Not Implemented',
			 502 => 'Bad Gateway',
			 503 => 'Service Unavailable',
			 504 => 'Gateway Time-Out',
			 505 => 'HTTP Version Not Supported',
			};
}

sub make_response {
  
  my $self = shift;
  my $code = shift;
  my $message = shift;
  
  my $response = new DataCash::XMLDocument();
  $response->set("status", $code);
  $response->set("reason", $message);
  return $response;

}


1;

#############################################################################

__END__

=head1 NAME

DataCash::XMLSSLAgent - Sends an XMLDocument to DataCash via HTTPS

=head1 SYNOPSIS

  my $agent = new DataCash::XMLSSLAgent(configfile => "/home/user/datacash.conf");
  my $response = $agent->send($xmldoc);

=head1 DESCRIPTION

An object-orientated module which is used to send a DataCash::XMLDocument 
object to the DataCash servers.

=head1 PUBLIC METHODS

=head2 new()

The constructor.  Requires a <code>configfile</code> parameter which is the
path to the XML configuration file.

=head2 send()
  
Sends the transaction to the DataCash servers.  Takes one parameter which 
is the DataCash::XMLDocument object.  Returns a DataCash::XMLDocument object
which contains the response from DataCash.

=head1 AUTHOR

  * Kas Robson, DataCash Ltd
  * Kenny Scott, DataCash Ltd


